//////////////////////////////////////////////////////////////////////////////
// Module	: ClsSmad.cpp												//
// Function	: Base class for all Sensoray Smart A/D boards.					//
// Author	: Jim Lamberson													//
// Created	: 08/03/1998													//
//////////////////////////////////////////////////////////////////////////////

#include <math.h>

#include "ClsSmad.h"		// Class declaration.

// Linearization table support --------------------------------------------

typedef struct {				// LINEARIZATION LOOKUP TABLE
	double	Slope;				//  Scalar to map sensor data into table index.
	double	Offset;				//  Offset to map sensor data into table index.
	double	Dat[512];			//  Table data to convert to engineering units.
} LOOKUP;

typedef struct {				// SENSOR ATTRIBUTES
	double	Scalar;				//  Scalar used to convert raw binary data into engineering units.
	LOOKUP	*LinTable;			//  Ptr to linearization LOOKUP table (if non-zero).
} SENSOR;

// Client-side linearization lookup tables.
#include "Typ1kNi2.h"		//   1K Ohm Nickel RTD
#include "Typ1kPt.h"		//   1K Ohm Platinum RTD

//////////////////////////////////////////////////////////////////////////////
// Linearization lookup table.
//////////////////////////////////////////////////////////////////////////////

#define typNop	( (LOOKUP *) 0 )	// Unimplemented linearization lookup table.

static const SENSOR SensorAttr[] =
{ //   Data   Ptr to
  //  Scalar  LinTbl
	{ 0.0010, typNop },		//  0: 10V **
	{ 0.0005, typNop },		//  1: 10V
	{ 0.0005, typNop },		//  2: 5V **
	{ 0.0002, typNop },		//  3: 5V
	{ 0.0001, typNop },		//  4: 1.65V **
	{ 0.0200, typNop },		//  5: 500mV
	{ 0.0050, typNop },		//  6: 100mV
	{ 0.0100, typNop },		//  7: 80mV **
	{ 0.1000, typNop },		//  8: E
	{ 0.1100, typNop },		//  9: J **
	{ 0.1000, typNop },		// 10: J
	{ 0.1700, typNop },		// 11: K **
	{ 0.1000, typNop },		// 12: K
	{ 0.1500, typNop },		// 13: T **
	{ 0.1000, typNop },		// 14: T 
	{ 0.6000, typNop },		// 15: S **
	{ 0.1000, typNop },		// 16: S
	{ 0.5000, typNop },		// 17: R **
	{ 0.1000, typNop },		// 18: R
	{ 0.1000, typNop },		// 19: 385 **
	{ 0.0500, typNop },		// 20: 385 platinum
	{ 0.1000, typNop },		// 21: 392 **
	{ 0.0500, typNop },		// 22: 392 platinum
	{ 0.0200, typNop },		// 23: 10K **
	{ 0.0100, typNop },		// 24: 10K
	{ 0.0100, typNop },		// 25: 4-20 mA
	{ 1.0000, typNop },		// 26: Strain/pressure gage
	{ 0.0200, typNop },		// 27: 400 ohm range
	{ 0.1250, typNop },		// 28: 4K ohm range
	{31.0000, typNop },		// 29: 600K ohm range
	{ 1.0000, typNop },		// 30: Custom resistive sensor
	{ 1.0000, typNop },		// 31: Disabled: remove from scan list
	{ 0.1000, typNop },		// 32: N
	{ 0.1000, typNop },		// 33: C
	{ 0.1000, typNop },		// 34: B
	{ 0.0100, typNop },		// 35: 120 nickel RTD
	{ 0.0500, typNop },		// 36: 200 nickel RTD
	{ 0.0100, typNop },		// 37: 1K nickel RTD
	{ 0.0125, typNop },		// 38: 385 platinum, HiRes (0.0125C)
	{ 0.0125, typNop },		// 39: 392 platinum, HiRes (0.0125C)
	{ 0.0100, typNop },		// 40: 10K on fixed 600K ohm range
	{ 0.1000, typNop },		// 41: 10 ohm
	// All of the following sensor types are processed as follows:
	//  1. Sensor type is configured as a "fundamental" type, such as V or Ohms.
	//  2. Linearization is performed by the Smart A/D class, rather than on board.
	{ 0.1250, &Typ1kNi2 },	// 42: 1K nickel RTD [4Kohm range]
	{ 0.1250, &Typ1kPt }	// 43: 1K platinum RTD [4Kohm range]
};

/////////////////////////////////////////////////////////////////////////////////////////
// Default error handler.  This function, which is not a class member, serves only as
// a place holder for an optional, external error handler.

static void DfltErrHandler( long errflags )
{
	// Nop.
}

//////////////////////////////////////////////////////////////////////////////
// Constructor.

ClsSmad::ClsSmad( short ProductID, long Adrs, short NumCh, const char *pSdcTbl, const char *pCmdTbl, char NumBoards )
{
	// Initialize board-level variables.
	MyFaultFlags	= 0;				// No faults have been detected so far.
	MyProductID		= ProductID;		// Save declared model number.
	MyAddress		= Adrs;				// Hardware address.
	MyNumChans		= NumCh;			// Highest valid channel number.
	MyNumGroups		= NumCh >> 3;		// Number of 8-channel groups on this model.
	MyNumBoards		= NumBoards;		// Number of Smart A/D boards residing on product.
	PhysSDC			= pSdcTbl;			// Pointer to Sensor definition code Logical-Physical mapping table.
	Cmd				= pCmdTbl;			// Pointer to Command opcodes Logical-Physical mapping table.
	pMyErrHandler	= DfltErrHandler;	// No app error handler has been declared yet.

	// Reset board's state image.
	ResetBoardProperties();
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Return true if channel number is valid, otherwise return false and call external error handler.

bool ClsSmad::IsValidChan( short Chan )
{
	// Return true if channel number is valid.
	if ( Chan < MyNumChans )
		return true;

	// Call external error handler if one has been declared.
	pMyErrHandler( ERR_BADPARM );

	// Indicate invalid channel number.
	return false;
}

//////////////////////////////////////////////////////////////////////////////
// Overloaded command handshake functions.

void ClsSmad::SendWord( short val )
{
	SendByte( HIBYTE(val) );
	SendByte( LOBYTE(val) );
}

void ClsSmad::SendWordRev( short val )
{
	SendByte( LOBYTE(val) );
	SendByte( HIBYTE(val) );
}

//////////////////////////////////////////////////////////////////////////////
// Status functions.

// Return Smart A/D board's address in the format appropriate for product type.
long ClsSmad::BoardAdrs()
{
	return MyAddress;
}

// Return standardized status flags (CRMT, DAV, ALARM, FAULT).
short ClsSmad::StatusFlagsGet()
{
	return (short)CmdStatusFlagsGet() | MyBoardState;
}

//////////////////////////////////////////////////////////////////////////////
// Error functions.

// Return all recorded error flags.
long ClsSmad::FaultFlagsGet()
{
	return MyFaultFlags;
}

// Set pointer to external, application-specific error handler callback, or to the default
// internal error handler if no external handler is specified (i.e., specifed as null pointer).
void ClsSmad::SetErrCallback( FPTR pErrHandler )
{
	pMyErrHandler = ( pErrHandler != 0 ) ? pErrHandler : DfltErrHandler;
}

// Declare one or more errors.
void ClsSmad::DeclareError( long ErrFlags )
{
	// Determine which flags are new, vs. which were already recorded.
	long newflags = ErrFlags & ~MyFaultFlags;

	// Record new error flags by combining with previously recorded error flags.
	MyFaultFlags |= ErrFlags;

	// Invoke application's error handler callback using only the new error flags.
	pMyErrHandler( newflags );
}

//////////////////////////////////////////////////////////////////////////////
// Time interval measurement functions.

// Return the time elapsed since StartTime.
double ClsSmad::ElapsedTime( double StartTime )						
{ 
	// Compute elapsed time.
	double t = SysTimer() - StartTime;

	// Compensate for clock overflow and return elapsed time.
	return ( t < -0.00001 ) ? ( t + 86400.0 ) : t;
}

//////////////////////////////////////////////////////////////////////////////
// Limit real value and convert to 16-bit signed integer.

short ClsSmad::LimTo32K( double Value )
{
	if		( Value >  32767.0 )	return  32767;
	else if ( Value < -32768.0 )	return -32768;
	else							return (short)Value;
}

///////////////////////////////////////////////////////////////////////////
// Set alarm limit values.

void ClsSmad::LimitValuesSet( short Chan, double LoVal, double HiVal )
{
	if ( IsValidChan( Chan ) )
	{
		register CHSTRUCT *c = &Ch[Chan];
		register double Scalar = SensorAttr[c->LogDefn].Scalar;
		c->LoLim = LimTo32K( LoVal / Scalar );
		c->HiLim = LimTo32K( HiVal / Scalar );
	}
}

//////////////////////////////////////////////////////////////////////////////
// Convert sensor data from Smart A/D into engineering units.
// New sensor types may be implemented either by this function (with the
// appropriate linearization processing) or by upgrading firmware on the
// Smart A/D board.

double ClsSmad::Linearize( short LSDC, short RawDat )
{
	LOOKUP *tbl = SensorAttr[LSDC].LinTable;
	
	if ( tbl )		// Linearization performed by Smart A/D class:
	{
		// Linearize Smart A/D channel data using 512-element lookup table ------------
		
		// Translate Value so that it falls within range of lookup table Dat[] values.
		register short N = (short)( (double)RawDat * tbl->Slope + tbl->Offset );
		if ( N < 0 )
			N = 0;
		else if ( N > 0x7FBFL )
			N = 0x7FBFL;

		// Set X such that the translated Value, N, is bounded by Dat[X] and Dat[X+1].
		long X = N >> 6;

		// Get linearization boundary values from table, interpolate between them, and
		// return with interpolated engineering units value.
		return tbl->Dat[X] + ( (tbl->Dat[X+1] - tbl->Dat[X]) * (double)(N & 0x3FL) ) / 64.0;
	}
	else			// Linearization performed by Smart A/D board:
		return (double)RawDat * SensorAttr[LSDC].Scalar;	//   Fetch sensor data & cvt to engr units
}

//////////////////////////////////////////////////////////////////////////////
// Read all channels from Smart A/D board into DatQ[].

void ClsSmad::GetAllRawData()
{
	// Get raw data from all channels into DatQ[].
	CmdOpen();
	for ( short i = 0; i < MyNumGroups ; i++ )
	{
		short g = i;
		SelectBoardByGroup( &g );
		SendByte( Cmd[CMD_READ8] | g );
		DeclareResponse( 16 );
	}
	CmdClose();
}

/////////////////////////////////////////////////////////////////////////////////
// Read all channels (scaled integer values)  from Smart A/D board into pData[].

void ClsSmad::RawDataAllGet( short *pData )
{
	// Get scaled integer sensor data from all channels into DatQ[].
	GetAllRawData();

	// Copy raw data to external application buffer.
	for ( register int i = 0; i < MyNumChans ; i++ )
		*pData++ = ReadWord();
}

//////////////////////////////////////////////////////////////////////////////
// Read all channels from Smart A/D board into pData[].

void ClsSmad::SensorsAllGet( double *pData )
{
	// Get scaled integer sensor data from all channels into DatQ[].
	GetAllRawData();

	// Convert raw data to engineering units and copy to external application buffer.
	CHSTRUCT *pC = &Ch[0];
	for ( register int i = 0; i < MyNumChans ; i++ )
		*pData++ = Linearize( (pC++)->LogDefn, ReadWord() );
}

//////////////////////////////////////////////////////////////////////////////
// Read one channel from Smart A/D board.

double ClsSmad::SensorGet( short Chan )
{
	return IsValidChan( Chan ) ? Linearize( Ch[Chan].LogDefn, RawChanGet( Chan ) ) : 0.0;
}

//////////////////////////////////////////////////////////////////////////////
// Read ambient temperature in degrees C from the specified TB.

double ClsSmad::AmbientTempGet( short Group )
{
	CmdOpen();
	SelectBoardByGroup( &Group );
	SendByte( Cmd[CMD_RDTEMP] | Group );
	DeclareResponse(2);
	CmdClose();
	return (double)ReadWord() * 0.1;
}

//////////////////////////////////////////////////////////////////////////////
// Calibrate Smart A/D input range.  Return calibrated EEPROM standard value.
// Range specifies input range (depends on product), while Value specifies
// the desired sensor data value at the applied input condition.

#define CAL_LM335	3		// Range = TC ref junction compensation sensor

double ClsSmad::Calibrate( short Chan, char Range, double val )
{
	if ( IsValidChan( Chan ) )
	{
		short CalValue = LimTo32K( val / SensorAttr[ Ch[Chan].LogDefn ].Scalar );
		CmdOpen();
		SelectBoardByChan( &Chan );
		SendByte( Cmd[CMD_CAL] | Chan );
		SendByte( Range );
		SendWord( CalValue );
		
		if ( Range == CAL_LM335 )
		{
			DeclareResponse(2);
			CmdClose();
			return (double)ReadWordRev();
		}
		else
		{
			DeclareResponse(4);
			CmdClose();
			return ReadReal();
		}
	}
	else
		return 0.0;
}

//////////////////////////////////////////////////////////////////////////////
// Return channel's logical sensor type.

short ClsSmad::SensorTypeGet( short Chan )
{
	return IsValidChan( Chan ) ? Ch[Chan].LogDefn : 0;
}

//////////////////////////////////////////////////////////////////////////////
// Program Open-Sensor failure mode for specified channel.

void ClsSmad::FailModeSet( short Chan, bool FailHi )
{
	if ( IsValidChan( Chan ) )
	{
		// Construct image of OpenType flags byte.
		short ChFirst = Chan & 0xF8;
		short ChLast = ChFirst | 7;
		char OpenFlags = 0;
		for ( int c = ChFirst; c <= ChLast; c++ )
			OpenFlags |= ( ( Ch[c].FailHigh = (FailHi != 0) ) << (c & 7) );

		// Set OpenType flags for one bank of 8 channels.
		short g = Chan >> 3;
		CmdOpen();
		SelectBoardByGroup( &g );
		SendByte( Cmd[CMD_DFNOPN] | g );
		SendByte( OpenFlags );
		CmdClose();
	}
}

bool ClsSmad::FailModeGet( short Chan )
{
	return IsValidChan( Chan ) ? Ch[Chan].FailHigh : false;
}

///////////////////////////////////////////////////////////////////////////
// Switch to high-speed, low resolution mode.

void ClsSmad::HiSpeedModeSet()
{
	if ( !MyIsModeHighSpeed )
	{
		CmdOpen();

		for ( short b = 0; b < MyNumBoards; b++ )
		{
			SelectBoard( b );
			SendByte( Cmd[CMD_DIAGNOSE] );
			SendByte( DIAG_HISPEEDMODE );
			SendByte( 0 );
		}

		CmdClose();
		MyBoardState |= STAT_HISPEED;
		MyIsModeHighSpeed = true;
	}
}

bool ClsSmad::HiSpeedModeGet()
{
	return MyIsModeHighSpeed;
}

///////////////////////////////////////////////////////////////////////////
// Return/set enable for high alarm limit detection.

bool ClsSmad::LimitHiEnabGet(short Chan)
{
	return IsValidChan( Chan ) ? Ch[Chan].HiLimEnable : false;
}

void ClsSmad::LimitHiEnabSet( short Chan, bool Enable )
{
	if ( IsValidChan( Chan ) )
	{
		// Cache the new Limit Enable state.
		CHSTRUCT *C = &Ch[Chan];
		C->HiLimEnable = Enable;

		// Send the command to the Smart A/D board.
		CmdOpen();
		SelectBoardByChan( &Chan );
		SendByte( Cmd[CMD_DFNLIM] | Chan );
		SendWord( C->HiLimEnable ? C->HiLim :  32767 );
		SendWord( C->LoLimEnable ? C->LoLim : -32768 );
		CmdClose();
	}
}

///////////////////////////////////////////////////////////////////////////
// Return/set enable for low alarm limit detection.

bool ClsSmad::LimitLoEnabGet(short Chan)
{
	return IsValidChan( Chan ) ? Ch[Chan].LoLimEnable : false;
}

void ClsSmad::LimitLoEnabSet( short Chan, bool Enable )
{
	if ( IsValidChan( Chan ) )
	{
		// Cache the new Limit Enable state.
		CHSTRUCT *C = &Ch[Chan];
		C->LoLimEnable = Enable;

		// Send the command to the Smart A/D board.
		CmdOpen();
		SelectBoardByChan( &Chan );
		SendByte( Cmd[CMD_DFNLIM] | Chan );
		SendWord( C->HiLimEnable ? C->HiLim :  32767 );
		SendWord( C->LoLimEnable ? C->LoLim : -32768 );
		CmdClose();
	}
}

///////////////////////////////////////////////////////////////////////////
// Switch to/from low-power standby mode.

void ClsSmad::LoPwrStandbySet( bool Standby )
{
	CmdOpen();

	for ( short b = 0; b < MyNumBoards; b++ )
	{
		SelectBoard( b );
		SendByte( Cmd[Standby ? CMD_SLEEP : CMD_WAKEUP] );
	}

	CmdClose();

	MyBoardState = Standby ? ( MyBoardState | STAT_SLEEPING ) : ( MyBoardState & ~STAT_SLEEPING);
}

///////////////////////////////////////////////////////////////////////////
// Return raw data from one channel.

short ClsSmad::RawChanGet( short Chan )
{
	if ( IsValidChan( Chan ) )
	{
		CmdOpen();
		SelectBoardByChan( &Chan );
		SendByte( (char)Chan );
		DeclareResponse(2);
		CmdClose();
		return ReadWord();
	}
	else
		return 0;
}

///////////////////////////////////////////////////////////////////////////
// Execute diagnostic function, copy response bytes to ByteBuf.

void ClsSmad::Diagnostic( short Chan, char DiagID, char Parm, char RespSize, char *pBuf )
{
	if ( IsValidChan( Chan ) )
	{
		// Invoke command to Smart A/D board and fetch the response.
		CmdOpen();
		SelectBoardByChan( &Chan );
		SendByte( Cmd[CMD_DIAGNOSE] | Chan );
		SendByte( DiagID );
		SendByte( Parm );
		DeclareResponse( RespSize );
		CmdClose();

		// Copy response to ByteBuf[].
		for ( int i = 0; i < RespSize; i++ )
			*pBuf++ = ReadByte();
	}
}

///////////////////////////////////////////////////////////////////////////
// Switch to 50 Hz rejection mode.

void ClsSmad::Reject50HzSet()
{
	if ( !( MyBoardState & STAT_REJECT50HZ ) )
	{
		CmdOpen();

		for ( short b = 0; b < MyNumBoards; b++ )
		{
			SelectBoard( b );
			SendByte( Cmd[CMD_FILT50] );
		}

		CmdClose();
		MyBoardState |= STAT_REJECT50HZ;
		MyIsModeReject50Hz = true;
	}
}

bool ClsSmad::Reject50HzGet()
{
	return MyIsModeReject50Hz;
}

/////////////////////////////////////////////////////////////////////////////////////////////
// Fetch firmware version number, x100, from Smart A/D board.  If more than one sub-board
// resides on the product, the first sub-board's version is returned.

short ClsSmad::VersionGet()
{
	CmdOpen();
	SelectBoard( 0 );
	SendByte( Cmd[CMD_DIAGNOSE] );
	SendByte( DIAG_GETVERSION );
	SendByte( 0 );
	DeclareResponse(2);
	CmdClose();
	
	return ReadWord();
}

///////////////////////////////////////////////////////////////////////////////
// Reset board and set FaultStuck flag if fault won't clear within one second.

void ClsSmad::ResetBoardProperties()
{
	// Indicate no pending board faults.
	MyFaultFlags &= ~( FAULT_DAV | FAULT_CRMT | FAULT_LED );

	// Initialize all board properties to reflect Reset state.
	MyBoardState		= 0;
	MyIsModeHighSpeed	= false;
	MyIsModeReject50Hz	= false;
	MyNetTimeout		= 10000;

	// Initialize all Channel properties for default sensor type.
	for ( short i = 0; i < MyNumChans; i++ )
		InitChanProperties( i, TYP_OLD_VOLTS_5V );
}

void ClsSmad::ResetBoard()
{
	// Initialize all board properties to reflect Reset state.
	ResetBoardProperties();
	
	// Invoke physical board hardware reset and validate status register signature.
	CmdResetBoard();
}

///////////////////////////////////////////////////////////////////////////
// Initialize Channel properties based on new logical SDC.

void ClsSmad::InitChanProperties( short Chan, short LogSDC )
{
	if ( IsValidChan( Chan ) )
	{
		if ( LogSDC >= 0 )
		{
			CHSTRUCT *c		= &Ch[Chan];	// INIT CHANNEL:
			c->LogDefn		= LogSDC;		//  logical sensor type code.
			c->HiLim		= 32767;		//  high alarm limit value.
			c->LoLim		= -32768;		//  low alarm limit value.
			c->Filter		= 0;			//  filter factor.
			c->FailHigh		= true;			//  fail type.
			c->LoLimEnable	= false;		//  low alarm enable.
			c->HiLimEnable	= false;		//  high alarm enable.
			c->LoAlarm		= false;		//  low alarm capture.
			c->HiAlarm		= false;		//  high alarm capture.
		}
	}
}

///////////////////////////////////////////////////////////////////////////
// Fetch word value from command response buffer.

short ClsSmad::ReadWord()
{
	short val = (unsigned char)*pDatQ++ << 8;
	return val | (unsigned char)*pDatQ++;
}

///////////////////////////////////////////////////////////////////////////
// Fetch word value, in reverse byte order, from command response buffer.

short ClsSmad::ReadWordRev()
{
	short val = (unsigned char)*pDatQ++;
	return val | ( (unsigned char)*pDatQ++ << 8 );
}

///////////////////////////////////////////////////////////////////////////
// Fetch real value from command response buffer.

double ClsSmad::ReadReal()
{
	// Handshake 4 bytes from Smart A/D board.
	long m0 = (unsigned char)*pDatQ++;
	long m1 = (unsigned char)*pDatQ++;
	long m2 = (unsigned char)*pDatQ++;
	int  xp = (unsigned char)*pDatQ++;
 
	if ( (xp < 28) || ((xp | m0 | m1 | m2) == 0) )
		return 0.0;

	double Sign = (m2 > 127) ? -1.0 : 1.0;

	if (xp > 228)
		return Sign * 1.33E+36;
	else
		return Sign * (double)( ( ( ( (m2 | 0x80L) << 8 ) + m1 ) << 8 ) + m0 ) * pow( 2.0, (double)(xp - 152) );
}

///////////////////////////////////////////////////////////////////////////
// Handshake a real onto a Smart A/D board.

void ClsSmad::SendReal( double N )
{
	int		exponent;
	double	mantissa = frexp( fabs(N), &exponent );

	if (exponent < -127)
	{
		SendWord( 0 );
		SendWord( 0 );
	}
	else
	{
		SendByte( (char)( mantissa * 16777216.0 ) );
		SendByte( (char)( mantissa *    65536.0 ) );
		SendByte( (char)( N > 0 ? mantissa * 256.0 - 128.0 : mantissa * 256.0 ) );
		SendByte( (char)( !N ? 0 : exponent + 128.0 ) );
	}
}

///////////////////////////////////////////////////////////////////////////
// Read & process limit violations from Smart A/D board.  Fill AlarmList[] with
// a list of all sounding channel numbers (HighAlarm=Chan, LowAlarm=Chan+0x10000).
// Return number of sounding channel alarms.

short ClsSmad::LimitAlarmsGet( short *AlarmList )
{
	short g;
	short i;
	
	// Reset channel alarm count.
	short nAlarms = 0;

	// Quickly test master alarm flag to determine if any channel alarms are sounding.
	// If alarm(s) are sounding, read all alarm flags & post to ChanStructs & AlarmList[].
	if ( ActiveALARM() )
	{
		// Begin new command cluster.
		CmdOpen();

		// For each 8-channel group ...
		for ( i = 0; i < MyNumGroups; i++ )
		{
			// Request that the group's alarm flags be copied into DatQ[].
			g = i;
			SelectBoardByGroup( &g );
			SendByte( Cmd[CMD_RDALRM] | g );
			DeclareResponse( 2 );
		}

		// Execute command cluster.
		CmdClose();

		// For each 8-channel group ...
		for ( i = 0; i < MyNumGroups; i++ )
		{
			// Fetch alarm flags from the group.
			char HiFlags = ReadByte();
			char LoFlags = ReadByte();

			// Init alarm flags channel number and mask.
			g = i << 3;
			char BitMask = 1;

			// For each channel in the group ...
			for ( short Chan = g; Chan < (g + 8); Chan++ )
			{
				// Set pointer to channel's attributes structure.
				CHSTRUCT *c = &Ch[Chan];

				// If high alarm is sounding ...
				if ( c->HiAlarm = ( ( HiFlags & BitMask ) != 0 ) )
				{
					c->HiLimEnable = false;
					AlarmList[nAlarms++] = Chan;
				}

				// If low alarm is sounding ...
				if ( c->LoAlarm = ( ( LoFlags & BitMask ) != 0 ) )
				{
					c->LoLimEnable = false;
					AlarmList[nAlarms++] = Chan | 0x80;
				}

				// Rotate channel mask to next channel.
				BitMask <<= 1;
			}
		}
	}
	
	return nAlarms;
}

//////////////////////////////////////////////////////////////////////////////
// Set channel filter factor.

void ClsSmad::FilterSet( short Chan, char Val )
{
	if ( IsValidChan( Chan ) )
	{
		// Cache the new filter factor.
		Ch[Chan].Filter = Val;

		// Issue SetFilter command to Smart A/D board.
		CmdOpen();
		SelectBoardByChan( &Chan );
		SendByte( Cmd[CMD_STFILT] | Chan );
		SendByte( Val );
		CmdClose();
	}
}

char ClsSmad::FilterGet( short Chan )
{
	return IsValidChan( Chan ) ? Ch[Chan].Filter : 0;
}

