//////////////////////////////////////////////////////////////////////////////
//	Module		: ClsStd.cpp												//
//	Function	: Base class for STDbus Smart A/D boards.					//
//	Author		: Jim Lamberson												//
//////////////////////////////////////////////////////////////////////////////

#include "ClsStd.h"			// Class declaration.

/////////////////////////////////////////////////////////////////////////////////////
// Opcode and SDC mapping tables are identical for all STDbus products.

// Logical-to-Physical command opcode mapping table.
static const char CmdTbl[] = {
	(const char)0x00,	// CMD_RDCHAN	--	Get data from channel
	(const char)0x10,	// CMD_DFNTYP	--	Declare sensor type
	(const char)0x20,	// CMD_DFNLIM	--	Set alarm limits
	(const char)0x30,	// CMD_RDALRM	--	Get alarm flags for 8 channels
	(const char)0x40,	// CMD_RDTEMP	--	Get TB temperature(s)
	(const char)0x50,	// CMD_DFNOPN	--	Set open-sensor values
	(const char)0x70,	// CMD_TARE		--	Tare gage
	(const char)0x80,	// CMD_FILT50	--	Set 50Hz filter mode
	(const char)0x90,	// CMD_READ8	--	Get data from 8 channels
	(const char)0xFF,	//					 spare
	(const char)0xE0,	// CMD_CAL		--	Calibration function
	(const char)0xF0,	// CMD_DIAGNOSE	--	Diagnostic function
	(const char)0xFF,	// CMD_RDSLOP	--	Get gage slope/intercept
	(const char)0xFF,	// CMD_WTSLOP	--	Set gage slope/intercept
	(const char)0xFF,	// CMD_GZERO	--	Set gage zero
	(const char)0xFF,	// CMD_GSPAN	--	Set gage span
	(const char)0x60,	// CMD_STFILT	--	Set filter factor
	(const char)0x43,	// CMD_SLEEP	--	Switch to low-power standby
	(const char)0x42,	// CMD_WAKEUP	--	Wake from low-power standby
	(const char)0xFF	// CMD_WCOEFF	--	Set custom-R coefficients
};

// Logical-to-physical SDC mapping table.
static const char SdcTbl[] = {
	(const char)0x00,	//  0 : 10V **
	(const char)0x15,	//  1 : 10V
	(const char)0x00,	//  2 : 5V **
	(const char)0x15,	//  3 : 5V
	(const char)0x0E,	//  4 : 1.65V **
	(const char)0x16,	//  5 : 500mV
	(const char)0x17,	//  6 : 100mV
	(const char)0x0D,	//  7 : 80mV **
	(const char)0x01,	//  8 : E
	(const char)0x02,	//  9 : J **
	(const char)0x1B,	// 10 : J
	(const char)0x03,	// 11 : K **
	(const char)0x1C,	// 12 : K
	(const char)0x04,	// 13 : T **
	(const char)0x1D,	// 14 : T
	(const char)0x05,	// 15 : S **
	(const char)0x1E,	// 16 : S
	(const char)0x06,	// 17 : R **
	(const char)0x1F,	// 18 : R
	(const char)0x07,	// 19 : 385 **
	(const char)0x18,	// 20 : 385 platinum
	(const char)0x08,	// 21 : 392 **
	(const char)0x19,	// 22 : 392 platinum
	(const char)0x0B,	// 23 : 10K **
	(const char)0x1A,	// 24 : 10K
	(const char)0x11,	// 25 : 4-20 mA
	(const char)0x0F,	// 26 : Strain/pressure gage
	(const char)0x09,	// 27 : 400 ohm range
	(const char)0x0A,	// 28 : 3K ohm range
	(const char)0x20,	// 29 : 600K ohm range
	(const char)0x0C,	// 30 : Custom resistive sensor
	(const char)0x13,	// 31 : Disabled: remove from scan list
	(const char)0x22,	// 32 : N
	(const char)0x23,	// 33 : C
	(const char)0x24,	// 34 : B
	(const char)0xFF,	// 35 : 120 nickel RTD
	(const char)0x28,	// 36 : 200 nickel RTD
	(const char)0x29,	// 37 : 1K nickel RTD
	(const char)0x2A,	// 38 : 385 platinum, HiRes (0.0125C)
	(const char)0x2B,	// 39 : 392 platinum, HiRes (0.0125C)
	(const char)0x25,	// 40 : 10K thermistor on fixed 600K ohm range
	(const char)0x2C,	// 41 : 10 ohm copper RTD
	(const char)0x10,	// 42 : 1K nickel RTD (4K range)
	(const char)0x10	// 43 : 1K platinum RTD (4K range)
};

//////////////////////////////////////////////////////////////////////////////
// Constructor.

ClsStd::ClsStd( short ProductID, long Adrs, short NumCh ) : ClsSmad( ProductID, Adrs, NumCh, SdcTbl, CmdTbl )
{
	// Base class does all initialization.
}

//////////////////////////////////////////////////////////////////////////////
// Command/Data queue insertion/extraction functions.

void ClsStd::SendByte( char val )
{
	*pCmdQ++ = val;			// Append byte to command queue.
}

//////////////////////////////////////////////////////////////////////////////
// Declare gage zero-scale condition.

void ClsStd::GageZeroSet( short Chan )
{
	GageZero = IsValidChan( Chan ) ? SensorGet( Chan ) : 0;
}

//////////////////////////////////////////////////////////////////////////////
// Set gage "span reference" for specified channel.

void ClsStd::GageSpanSet( short Chan, short GageLoad )
{
	if ( IsValidChan( Chan ) )
	{
		// Simulate GageSpanSet command.
		CHSTRUCT *C = &Ch[Chan];
		double GageSpan = SensorGet( Chan );
		double NewParm = (C->GagemVperV * C->GageMaxLoad) * (GageLoad / (GageSpan - GageZero));
		C->GagemVperV = (short)(1 + NewParm / 32768.0);					// mV/V (x10)
		C->GageMaxLoad = (short)(NewParm / (double)C->GagemVperV);		// max pressure

		// Redeclare sensor type as gage to set new gage parameters in effect.
		CmdOpen();
		SendByte( Cmd[CMD_DFNTYP] | Chan );
		SendByte( PhysSDC[C->LogDefn] );
		SendWord( C->GagemVperV );
		SendWord( C->GageMaxLoad );
		SendWord( 350 );

		// Set current gage load value to specified full load.
		SendByte( Cmd[CMD_TARE] | Chan );
		SendWord( GageLoad );
		CmdClose();
	}
}

//////////////////////////////////////////////////////////////////////////////
// Tare gage.

void ClsStd::GageTareSet( short Chan )
{
	if ( IsValidChan( Chan ) )
	{
		CmdOpen();
		SendByte( Cmd[CMD_TARE] | Chan );
		SendWord( 0 );
		CmdClose();
	}
}

//////////////////////////////////////////////////////////////////////////////
// Set linearization coefficients for custom resistive sensor channel.

void ClsStd::CoefficientsSet( short Chan, double c1, double c2, double c3 )
{
	if ( IsValidChan( Chan ) )
	{
		CmdOpen();
		InitChanProperties( Chan, TYP_OHMS_CUSTOM );
		SendByte( Cmd[CMD_DFNTYP] | Chan );
		SendByte( PhysSDC[TYP_OHMS_CUSTOM] );
		SendWord( LimTo32K(c1) );
		SendWord( LimTo32K(c2) );
		SendWord( LimTo32K(c3) );
		CmdClose();
	}
}

///////////////////////////////////////////////////////////////////////////
// Program sensor type.

void ClsStd::SensorTypeSet( short Chan, short LogSDC )
{
	if ( IsValidChan( Chan ) )
	{
		// If sensor type is in legal range ...
		if ( LogSDC < NUM_SDCS )
		{
			// If sensor type is supported by this product ...
			if ( PhysSDC[LogSDC] != 0xFF )
			{
				short LogChan = Chan;
				InitChanProperties( LogChan, LogSDC );

				// Declare the new sensor type to the Smart A/D board -----------

				CmdOpen();									// Begin new command.

				SendByte( Cmd[CMD_DFNTYP] | Chan );			// Issue SetSensorType command.
				SendByte( PhysSDC[LogSDC] );

				if ( LogSDC == TYP_OHMS_CUSTOM )			// Append additional parameters, if any.
				{
					SendWord( 0 );
					SendWord( 1 );
					SendWord( 0 );
				}
				else if ( LogSDC == TYP_GAGE )
				{
					SendWord( Ch[Chan].GagemVperV = 28 );
					SendWord( Ch[Chan].GageMaxLoad = 10000 );
					SendWord( 350 );
				}

				CmdClose();									// Execute the command.

				return;										// Exit.
			}
		}

		// Sensor type is not supported or is outside legal value range, so call external error handler.
		pMyErrHandler( ERR_BADPARM );
	}
}

///////////////////////////////////////////////////////////////////////////////
// Direct I/O functions.

char ClsStd::InByteDirect(long Adrs)
{
	return InByte( Adrs );
}

void ClsStd::OutByteDirect(long Adrs, char Value)
{
	OutByte( Adrs, Value );
}

bool ClsStd::ActiveALARM()
{
	return ( InByte( MyAddress + 1 ) & STAT_ALARM ) != 0;
}

unsigned char ClsStd::CmdStatusFlagsGet()
{
	return InByte( MyAddress + 1 ) & 0xF0;
}

////////////////////////////////////////////////////////////////////////////////////
// Wait for active handshake flag.

bool ClsStd::HandshakeWait( char mask, char xormask, double maxtime )
{
	// Log the start time.
	double StartTime = SysTimer();

	// Wait for handshake flag or timeout, whichever occurs first.
	do
	{
		if ( ( ( InByte( MyAddress + 1 ) & mask ) ^ xormask ) != 0 )
			return true;
	}
	while ( ElapsedTime( StartTime ) < maxtime );

	// Max time elapsed, so we check handshake flag one final time in case this function
	// was interrupted for a long time.
	return ( ( ( InByte( MyAddress + 1 ) & mask ) ^ xormask ) != 0 );
}

///////////////////////////////////////////////////////////////////////////////
// Reset board and set FaultStuck flag if fault won't clear within one second.

#define MAX_RESET_RETRIES	3

void ClsStd::CmdResetBoard()
{
	char	status;
	int		retry	= 0;

	// Repeat until the board's Fault flag is set or we reach max allowed retries ...
	while ( 1 )
	{
		// Invoke physical board hardware reset.
		OutByte( MyAddress + 1, 0 );

		// If the board's Fault flag is active, as it should be ...
		if ( InByte( MyAddress + 1 ) & STAT_FAULT )
		{
			// Wait for Fault to clear, or timeout, whichever occurs first.  If we timed out while waiting for Fault to clear ...
			if ( !HandshakeWait( STAT_FAULT, STAT_FAULT, 1.0 ) )
			{
				// Indicate problem with Fault: it won't switch to the inactive state.
				DeclareError( FAULT_LED );
			}
			else
			{
				// Cache the board's status byte.
				status = InByte( MyAddress + 1 );

				// Detect invalid DAV state.
				if ( ( status & STAT_DAV ) != 0 )
					DeclareError( FAULT_DAV );

				// Detect invalid CRMT state.
				if ( ( status & STAT_CRMT ) == 0 )
					DeclareError( FAULT_CRMT );
			}

			// Exit loop; we are done.
			break;
		}

		// Bump the retry counter.  If max allowed retries has been exceeded ...
		else if ( retry++ > MAX_RESET_RETRIES )
		{
			// Indicate problem with Fault: it won't switch to the active state.
			DeclareError( FAULT_LED );

			// Exit loop; we are done.
			break;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Handshake any pending command string from CmdQ[] to board.

void ClsStd::CmdClose()
{
	short	i;
	short	nCmdBytes;
	char	*pCmdByte;

	// If no CRMT fault previously detected ...
	if ( MyFaultFlags & FAULT_CRMT )
	{
		// Determine number of pending command bytes in CmdQ[], if any.
		nCmdBytes = (short)( pCmdQ - CmdQ );

		// Handshake all pending command bytes to board.
		pCmdByte = CmdQ;
		for ( i = 0; i < nCmdBytes; i++ )
		{
			// Wait for active CRMT or timeout.  If the wait timed out ...
			if ( !HandshakeWait( (const char)STAT_CRMT, 0, MAXHSTIME ) )
			{
				// Declare handshake fault and terminate command transmission.
				DeclareError( FAULT_CRMT );
				break;
			}

			// Send byte to Smart A/D board.
			OutByte( MyAddress, *pCmdByte++ );
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Fetch command response string from board into DatQ[].

void ClsStd::HandshakeDatString( short nBytes = 0 )
{
	short	i;

	// If no DAV fault was previously detected ...
	if ( ( MyFaultFlags & FAULT_DAV ) == 0 )
	{
		// Repeat for each expected response byte ...
		for ( i = 0; i < nBytes; i++ )
		{
			// Wait for active DAV or timeout.  If the wait timed out ...
			if ( !HandshakeWait( STAT_DAV, 0, MAXHSTIME ) )
			{
				// Declare handshake fault and terminate command transmission.
				DeclareError( FAULT_DAV );
				break;
			}

			// Fetch response byte and store in DatQ[].
			*pDatQ++ = InByte( MyAddress );
		}

		// Reset DatQ[] pointer so the first call to ReadByte() will return the proper value.
		pDatQ = DatQ;
	}
}

//////////////////////////////////////////////////////////////////////////////
// Send any pending command to board and fetch response string into DatQ[].

void ClsStd::DeclareResponse( short nBytes )
{
	// Send any pending command string in CmdQ[] to board.
	CmdClose();

	// Fetch response string from board into DatQ[].
	HandshakeDatString( nBytes );

	// Reset CmdQ[] pointer in preparation for next command string.
	pCmdQ = CmdQ;
}

///////////////////////////////////////////////////////////////////////////
// Request product model number and return expected response byte count.

short ClsStd::ProductIDGet()
{
	CmdOpen();
	SendByte( Cmd[CMD_DIAGNOSE] );
	SendByte( DIAG_GETPRODUCTID );
	SendByte( 0 );
	DeclareResponse( 2 );
	CmdClose();
	return ReadWord();
}
